
#ifndef _XSEC_H_
#define _XSEC_H_

/*
 * Header file for the X11 security extension.
 *
 * Copyright (c) 1994 SecureWare, Inc.  All Rights Reserved.
 *
 * @(#)05 1.8 x11r5/extensions/include/xsecurity/xsecurity.h, xsecext, sco_3.0, 301FCS 12:19:30 12/8/94, SecureWare, Inc.
 */

/* AIX compiler does not seem to like the cast of pointers on the lhs of an
 * expression so they were reworked to eliminate redundant casts.
 */

#include <sys/secdefines.h>

#if SEC_BASE /* { */

#include <X11/extensions/Xsecdefs.h>
#include <X.h>
#include <misc.h>
#include <extnsionst.h>
#include <input.h>
#include <cursor.h>
#include <property.h>
#include <resource.h>
#include <pixmap.h>
#include <window.h>
#include <colormap.h>

/* Object IDs for special audit objects (for floating) */

#define ClientObject		-1
#define	KeyboardObject		-2
#define	EventObject		-3

#define DrawableObjectType(d)	((d)->type == DRAWABLE_PIXMAP ? PixmapObject : \
								WindowObject)

/*
 * Internal object identifiers for standard objects.
 */
#define WindowObject		0
#define	PixmapObject		1
#define GCObject		2
#define	FontIDObject		3
#define	CursorObject		4
#define	PropertyObject		5
#define	AtomObject		6
#define	ColormapObject		7
#define	ColorCellObject		8
#define 	X_NUM_OBJECT_TYPES	9

/*
 * Attribute type identifiers.
 * These are in a continuous sequence with the object types above simply
 * so that we can use the same table to create the atoms and yet maintain
 * uniqueness.  The object types are used to index into the client's
 * permissions array so they should be first.
 */
#define		FirstClientAttrType	9
#define ClientLUIDAttr		 9
#define	ClientUIDAttr		10
#define	ClientGIDAttr		11
#define	MaxPrivAttr		12
#define	EffPrivAttr		13
#define	ProcessIDAttr		14
#define		LastClientAttrType	16

#define	MACLabelAttr		15	/* This is both client and obj attr */
#define InfoLabelAttr		16	/* This is both client and obj attr */

#define		FirstObjectAttrType	15
#define	InputInfoLabelAttr	17
#define	PermsAttr		18
#define	UIDAttr			19
#define	GIDAttr			20
#define		LastObjectAttrType	20

#define		X_NUM_CLIENT_ATTRS	(LastClientAttrType -		\
					 FirstClientAttrType + 1)

#define		X_NUM_OBJECT_ATTRS	(LastObjectAttrType -		\
					 FirstObjectAttrType + 1)

typedef unsigned long	DACMode;
typedef PrivWord	PrivSet[X_PRIVSET_SIZE];
typedef PrivWord	*PrivSetPtr;

/*
 * Subject security attributes - this structure is used to maintain
 * the security attributes associated with a client. The structure
 * is used in the ClientSecAttrs[] array but is associated with X
 * objects and other structures such as grab records and event lists.
 */

typedef struct _SubjSecAttr {
    int		refcnt;			/* Reference count */
    int		clientID;		/* Client table index */
    LoginUserID loginUserID;		/* Client LUID */
    UserID	userID;
    int		nGroups;		/* Size of group ID set */
    GroupID	*pGroupIDs;		/* Group ID set */
    DACMode	defaultPerms[X_NUM_OBJECT_TYPES];	/* Default DAC */
#if SEC_MAC
    MACLabel	*macLabel;		/* Client MAC label */
    MACLabel	*propMacLabel;		/* For property operations */
    unsigned	propCompose:1;		/* Boolean -- compose prop. atoms? */
    unsigned	propMultilevel:1;	/* Boolean -- special ML mode? */
#endif
#if SEC_ILB
    InfoLabel	*requestInfoLabel;	/* Request buffer IL */
    InfoLabel	*replyInfoLabel;	/* Reply/Event/Error buffer IL */
#endif
    PrivSet	maxPrivs;		/* Maximum privilege set */
    PrivSet	effPrivs;		/* Effective privilege set */
    ProcessID	processID;		/* For auditing */
} SubjSecAttr, *SubjSecAttrPtr;

/* Parallel client security attributes table */

extern SubjSecAttrPtr *clientSecAttrs;		/* XSecurityExtensionInit() */

#define ClientSecAttrs(cptr)		(clientSecAttrs[cptr->index])
#define	ClientSecAttrsByClient(cptr)	(clientSecAttrs[cptr->index])
#define ClientSecAttrsByID(cindex)	(clientSecAttrs[cindex])

#define ClientReqType(cptr)	(((xReq *)((cptr)->requestBuffer))->reqType)

#if SEC_MAC
#define cMacLabel(c)	   ((MACLabelPtr)(ClientSecAttrs(c)->macLabel))
#if SEC_ILB
#define	cReqInfoLabel(c)   ((InfoLabelPtr)(ClientSecAttrs(c)->requestInfoLabel))
#define cReplyInfoLabel(c) ((InfoLabelPtr)(ClientSecAttrs(c)->replyInfoLabel))
#endif
#endif

/*
 * Object security attributes - associates security attributes with X
 * objects.
 */

typedef struct _ObjSecAttr {
    void	*extra;			/* Miscellaneous storage */
    int		objType;		/* Which object type */
    int		creator;		/* Index of creator in clients table */
    UserID	userID;			/* Owning user ID */
    GroupID	groupID;		/* Owning group ID */
    DACMode	perms;			/* DAC Permissions */
#if SEC_MAC
    MACLabel	*macLabel;		/* MAC label */
#endif
#if SEC_ILB
    InfoLabel	*infoLabel;		/* Information label */
    InfoLabel	*inputInfoLabel;	/* Input information label (window) */
#endif
} ObjSecAttr, *ObjSecAttrPtr;

/* Miscellaneous structures for use with X objects */

typedef struct _RegService {
	struct _RegService	*next;
	XID		id;		/* XID associated with registration */
	WindowPtr	regWindow;	/* Registration Window */
	Atom		serviceName;	/* Name of service */
	Window		serviceWindow;	/* Associated window */
	Bool		exitFlag;	/* Server to exit if service goes? */
} RegServiceRec, *RegServicePtr;

typedef struct _OtherSecClients {
	struct _OtherSecClients	*next;
	XID			resource;	/* used with Resource Manager */
	WindowPtr		window;		/* associated window pointer */
	OtherClientsPtr		otherClient;	/* corresponding structure */
	Mask			mask;		/* standard event mask */
	SubjSecAttrPtr		eventAttrs;	/* event security attributes */
	Mask			secEventMask;	/* security events selected */
	SubjSecAttrPtr		secEventAttrs;	/* security event attributes */
} OtherSecClients, *OtherSecClientsPtr;

/* Macros for atom security attribute handling */

#define	atomSecAttrsAssign(n)	((n)->secAttrs)
#define	atomSecAttrs(n)		((ObjSecAttrPtr)atomSecAttrsAssign(n))
#define atomMacLabelAssign(n)	(atomSecAttrsAssign(n)->macLabel)
#define atomMacLabel(n)		((MACLabelPtr)atomSecAttrs(n)->macLabel)
#if SEC_ILB
#define	atomInfoLabel(n)	((InfoLabelPtr)(atomSecAttrs(n)->infoLabel))
#endif

/*
 * Window property structure used to maintain security attributes for each
 * property.
 */

typedef struct _PropertySecRec {
	struct _PropertySecRec	*next;
	PropertyPtr		prop;		/* corresponding property */
	ObjSecAttrPtr		secAttrs;	/* security attributes */
} PropertySecRec, *PropertySecRecPtr;

#if SEC_ILB
#define	propInfoLabel(p)	((InfoLabelPtr)((p)->secAttrs->infoLabel))
#endif

/*
 * Pixmap structure used to maintain the security attributes and the actual
 * structure pointer for the PixmapRec since the RT_PIXMAP resource ID yields
 * a pointer to this structure and not the PIXMAP.
 *
 * Note that the first element of the structure is an unsigned char field
 * which is filled in to DRAWABLE_PIXMAP when a PixmapSecRec is created.
 * This is designed to match the structure definitions for both the Window
 * and Pixmap structures which are cached as RC_DRAWABLE resources which
 * include a DrawableRec as the first member of each structure. The type
 * field is required by many macros to determine which type of object the
 * resulting resource value is associated with following a RC_DRAWABLE
 * lookup. See 'include/pixmapstr.h' for the DrawableRec structure which
 * can not be included here due to dependencies.
 */

typedef struct _PixmapSecRec {
	unsigned char	type;			/* type = DRAWABLE_PIXMAP */
	PixmapPtr	pPixmap;		/* actual PixmapRec pointer */
	ObjSecAttrPtr	secAttrs;		/* pixmap security attributes */
} PixmapSecRec, *PixmapSecRecPtr;

#if SEC_ILB
#define pmapInfoLabel(p)  ((InfoLabelPtr)((p) ? (p)->secAttrs->infoLabel : 0))
#endif

/*
 * Colorcell security attribute structures designed to parallel the data
 * structures used maintain the colorcells for the various visual types.
 * The SEC_LOCO structure is used for DirectColor visuals in the red, green,
 * and blue colormap channels. For all other visuals, the SEC_LOCO structure
 * is used in the red channel only. The SEC_SHCO structure is used for the
 * PseudoColor or GrayScale visuals when a client allocates pixels using the
 * AllocColorPlanes() request. This requires treating the colorcells as if
 * the visual was a DirectColor visual.
 */

typedef struct _SEC_LOCO {
	ObjSecAttrPtr	secAttrs;		/* colorcell security attrs */
} SEC_LOCO;

typedef struct _SEC_SHAREDCOLOR {
	ObjSecAttrPtr	secAttrs;		/* colorcell security attrs */
} SEC_SHAREDCOLOR, *SEC_SHAREDCOLOR_PTR;

typedef struct _SEC_SHCO {
	SEC_SHAREDCOLOR		*red;		/* red channel */
	SEC_SHAREDCOLOR		*green;		/* green channel */
	SEC_SHAREDCOLOR		*blue;		/* blue channel */
} SEC_SHCO;

typedef struct _ColorcellSecRec {
	union {
		SEC_LOCO	local;
		SEC_SHCO	shco;
	} co;
} ColorcellSecRec, *ColorcellSecRecPtr, **ColorcellSecRecList;

/* Colorcell security attribute macros */

#if SEC_ILB
#define	colorInfoLabel(a)	((InfoLabelPtr)((a)->infoLabel))
#endif

/*
 * Colormap structure used to maintain the security attributes and the actual
 * structure pointer for the ColormapRec since the RT_COLORMAP resource ID
 * yields a pointer to this structure and not the ColormapRec.
 */

typedef struct _ColormapSecRec {
	ColormapPtr		pMap;		/* actual ColormapRec pointer */
	ColorcellSecRec		*red;		/* red channel sec attrs */
	ColorcellSecRec		*green;		/* green channel sec attrs */
	ColorcellSecRec		*blue;		/* blue channel sec attrs */
	ObjSecAttrPtr		secAttrs;	/* colormap security attrs */
} ColormapSecRec, *ColormapSecRecPtr;

/*
 * Cursor structure used to maintain the security attributes and the actual
 * structure pointer for the CursorRec since the RT_CURSOR resource ID yields
 * a pointer to this structure and not the CursorRec.
 */

typedef struct _CursorSecRec {
	CursorPtr	pCursor;		/* actual CursorRec pointer */
	ObjSecAttrPtr	secAttrs;		/* cursor security attributes */
} CursorSecRec, *CursorSecRecPtr;

/*
 * Grab structure used to indicate whether an existing active grab or a
 * passive grab for a window is an override grab. Passive grabs have client
 * security attributes associated with them for privilege checks.
 */

typedef struct _GrabSecRec {
	struct _GrabSecRec	*next;
	GrabPtr			grab;		/* associated structure */
	WindowPtr		window;		/* associated window */
	SubjSecAttrPtr		grabAttrs;	/* security attributes */
	Bool			isOverride;	/* override grab? */
} GrabSecRec, *GrabSecRecPtr;

/* Macros and defines for device specific GrabSecRec reference */

#define KEYBOARD_DEVICE	0
#define	POINTER_DEVICE	1

extern GrabSecRec DeviceGrabSecRec[];

#define	kbdGrabPtr		(DeviceGrabSecRec[KEYBOARD_DEVICE].grab)
#define	kbdGrabIsOverride	(DeviceGrabSecRec[KEYBOARD_DEVICE].isOverride)
#define	kbdGrabSecAttrs		(DeviceGrabSecRec[KEYBOARD_DEVICE].grabAttrs)

#define	ptrGrabPtr		(DeviceGrabSecRec[POINTER_DEVICE].grab)
#define	ptrGrabIsOverride	(DeviceGrabSecRec[POINTER_DEVICE].isOverride)
#define	ptrGrabSecAttrs		(DeviceGrabSecRec[POINTER_DEVICE].grabAttrs)

/*
 * These are the data structures required to maintain security
 * specific information for different X objects without having
 * to modify the X (server/include) data structures directly
 * thereby impacting DDX binary compatibility.
 */

typedef struct _ScreenSecInfo {
    short		ssHeight;	/* height of security stripe */
    pointer		resColor;	/* reserved colorcell support */
#if SEC_ILB
    void		(*ClearToBackground)();		/* wrapper routine */
    void		(*PaintWindowBackground)();	/* wrapper routine */
#endif
#ifdef SCO
    Bool		(*SaveScreen)();		/* wrapper routine */
#endif
} ScreenSecInfo, *ScreenSecInfoPtr;

extern int	secExtScreenPrivateIndex;

#define sPrivateAssign(scr)	((scr)->devPrivates[secExtScreenPrivateIndex].ptr)
#define sPrivate(scr)		((ScreenSecInfoPtr)sPrivateAssign(scr))

#define	sHeight(scr)		(sPrivate(scr)->ssHeight)
#define sResColorAssign(scr)	(sPrivate(scr)->resColor)
#define sResColor(scr)		((ColorResPtr)sResColorAssign(scr))
#if SEC_ILB
#define	sClearBack(scr)		(sPrivate(scr)->ClearToBackground)
#define sPaintBack(scr)		(sPrivate(scr)->PaintWindowBackground)
#endif
#ifdef SCO
#define sSaveScreen(scr)	(sPrivate(scr)->SaveScreen)
#endif

typedef struct _WindowSecRec {
    XID			wid;		/* associated window resource ID */
    WindowPtr		window;		/* window data structure pointer */
    ObjSecAttrPtr	secAttrs;	/* object security attributes */
    SubjSecAttrPtr	eventAttrs;	/* subject attrs for event selection */
    Mask		secEventMask;	/* security event mask */
    SubjSecAttrPtr	secEventAttrs;	/* subject attrs for security event */
    Mask		otherSecEventMasks; /* aggregate security event mask */
    OtherSecClientsPtr	otherSecClients; /* list of security event selectors */
    RegServicePtr	serviceList;	/* service registration list */
    PropertySecRecPtr	propList;	/* window property security attrs */
    GrabSecRecPtr	grabList;	/* passive grab security attributes */
#if SEC_ILB
    InfoLabelPtr	explicitIL;	/* IL set using XSetObjectInfoLabel */
    InfoLabelPtr	contentsIL;	/* window data info label */
    InfoLabelPtr	attrsIL;	/* window attributes info label */
    InfoLabelPtr	borderPixmapIL;	/* border pixmap info label */
    int			backPixmapChg;	/* deferred background pixmap flag */
    InfoLabelPtr	backPixmapIL;	/* background pixmap info label */
    InfoLabelPtr	newBackPixmapIL; /* new background pixmap info label */
#endif
} WindowSecRec, *WindowSecRecPtr;

/* Convenience macros for security attribute access */

/*
 * Window-related macros.
 */

extern int	secExtWindowPrivateIndex;

#define	wSecPrivateAssign(w)		\
	((w)->devPrivates[secExtWindowPrivateIndex].ptr)
#define	wSecPrivate(w)		\
	((WindowSecRecPtr)wSecPrivateAssign(w))

#define wSecWid(w)		(wSecPrivate(w)->wid)
#define wSecWinPtr(w)		(wSecPrivate(w)->window)
#define	wSecAttrsAssign(w)	(wSecPrivate(w)->secAttrs)
#define	wSecAttrs(w)		((ObjSecAttrPtr)wSecPrivate(w)->secAttrs)
#define wEventAttrs(w)		(wSecPrivate(w)->eventAttrs)
#define	wSecEventMask(w)	(wSecPrivate(w)->secEventMask)
#define wSecEventAttrs(w)	(wSecPrivate(w)->secEventAttrs)
#define	wOtherSecEventMasks(w)	(wSecPrivate(w)->otherSecEventMasks)
#define wOtherSecClients(w)	(wSecPrivate(w)->otherSecClients)
#define	wServiceListAssign(w)	(wSecPrivate(w)->serviceList)
#define	wServiceList(w)		((RegServicePtr)wServiceListAssign(w))
#define wGrabSecListAssign(w)	(wSecPrivate(w)->grabList)
#define wGrabSecList(w)		((GrabSecRecPtr)wGrabSecListAssign(w))
#define wPropertySecListAssign(w)	(wSecPrivate(w)->propList)
#define wPropertySecList(w)		((PropertySecRecPtr)wPropertySecListAssign(w))

#define wMacAttrs(w)		(wSecAttrs(w)->extra ?		\
				    (ObjSecAttrPtr)(wSecAttrs(w)->extra) : \
				    wSecAttrs(w))

#define wMacLabel(w)		((MACLabelPtr)(wMacAttrs(w)->macLabel))
#if SEC_ILB
#define wInfoLabel(w)		((InfoLabelPtr)(wMacAttrs(w)->infoLabel))
#define wInputInfoLabel(w)	((InfoLabelPtr)(wMacAttrs(w)->inputInfoLabel))

#define wExplicitILAssign(w)		(wSecPrivate(w)->explicitIL)
#define wExplicitIL(w)			((InfoLabelPtr)wExplicitILAssign(w))
#define wContentsILAssign(w)		(wSecPrivate(w)->contentsIL)
#define wContentsIL(w)			((InfoLabelPtr)wContentsILAssign(w))
#define	wAttrsILAssign(w)		(wSecPrivate(w)->attrsIL)
#define	wAttrsIL(w)			((InfoLabelPtr)wAttrsILAssign(w))
#define wBorderPixmapILAssign(w)	(wSecPrivate(w)->borderPixmapIL)
#define wBorderPixmapIL(w)		((InfoLabelPtr)wBorderPixmapILAssign(w))
#define	wBackPixmapChgAssign(w)		(wSecPrivate(w)->backPixmapChg)
#define	wBackPixmapChg(w)		((int)wBackPixmapChgAssign(w))
#define wBackPixmapILAssign(w)		(wSecPrivate(w)->backPixmapIL)
#define	wBackPixmapIL(w)		((InfoLabelPtr)wBackPixmapILAssign(w))
#define wNewBackPixmapILAssign(w)	(wSecPrivate(w)->newBackPixmapIL)
#define wNewBackPixmapIL(w)		((InfoLabelPtr)wNewBackPixmapILAssign(w))
#endif

/* GC structure and macros */

typedef struct _GcSecRec {
	ObjSecAttrPtr	secAttrs;		/* GC security attributes */
#if SEC_ILB
	InfoLabelPtr	attrsIL;		/* GC attributes Il */
	InfoLabelPtr	tileIL;			/* Tile pixmap IL */
	InfoLabelPtr	stippleIL;		/* Stipple pixmap IL */
	InfoLabelPtr	clipMaskIL;		/* Clip Mask pixmap IL */
#endif
} GcSecRec, *GcSecRecPtr;

extern int	secExtGCPrivateIndex;

#define	gcSecPrivateAssign(gc)	((gc)->devPrivates[secExtGCPrivateIndex].ptr)
#define	gcSecPrivate(gc)	((GcSecRecPtr)gcSecPrivateAssign(gc))

#define	gcSecAttrsAssign(gc)	(gcSecPrivate(gc)->secAttrs)
#define	gcSecAttrs(gc)		((ObjSecAttrPtr)gcSecAttrsAssign(gc))
#if SEC_ILB
#define gcInfoLabelAssign(gc)	(gcSecAttrs(gc)->infoLabel)
#define gcInfoLabel(gc)		((InfoLabelPtr)gcInfoLabelAssign(gc))
#define gcAttrsILAssign(gc)	(gcSecPrivate(gc)->attrsIL)
#define gcAttrsIL(gc)		((InfoLabelPtr)gcAttrsILAssign(gc))
#define	gcTileILAssign(gc)	(gcSecPrivate(gc)->tileIL)
#define	gcTileIL(gc)		((InfoLabelPtr)gcTileILAssign(gc))
#define	gcStippleILAssign(gc)	(gcSecPrivate(gc)->stippleIL)
#define	gcStippleIL(gc)		((InfoLabelPtr)gcStippleILAssign(gc))
#define	gcClipMaskILAssign(gc)	(gcSecPrivate(gc)->clipMaskIL)
#define	gcClipMaskIL(gc)	((InfoLabelPtr)gcClipMaskILAssign(gc))
#endif

/* Miscellaneous macros */

#define CopySubjSecAttrs(attrp)		(++(attrp)->refcnt, (attrp))

#define	AddColorCellSecAttrs(ap, a)	AddLinkedSecAttrs(ap, a)
#define	FreeAllColorCellSecAttrs(ap)	FreeAllLinkedSecAttrs(ap)

/*
 * These are the permissions of the "server client" and, thus, of all
 * objects created by the server client at startup (root window, default
 * colormap, etc.).
 */

#define SERVER_DAC_MODE	    ((DACAll << DACUserShift) |			\
			    ((DACRead|DACQuery|DACInsert)		\
				<< DACGroupShift) |			\
			    ((DACRead|DACQuery|DACInsert)		\
				<< DACOtherShift))

/*
 * Multilevel (polyinstantiated) property-related defines.
 */

#define COMPOSE_PREFIX		"__MLPROP_%lu"
#define COMPOSE_SEPARATOR	'.'
#define COMPOSE_BYTE		"%x"

/*
 * Security atom translations.
 */
typedef struct _sec_atom {
    char	*name;		/* Atom string */
    int		id;		/* Internal ident. to which the atom maps */
    Atom	atom;		/* Atom number assigned at runtime */
} SecAtom;

/*
 * Event cleaner (used for zero'ing unused portions of events to prevent
 * object reuse booboos).
 */
typedef struct _ev_cleaner {
    int		size;		/* Beyond this can always be 0'd */
    int		gapoff[10];	/* Up to 10 1-byte gaps can be 0'd */
} EventCleaner;

/*
 * The information concerning how many colors are reserved and which
 * colormap contains the reservations rightfully belongs in the VisualRec
 * structure.  However, this has much impact on the DDX layer and we're
 * trying to stay out of that as much as possible.
 */
typedef struct _ColorRes {
    struct _ColorRes *next;
    unsigned long vid;			/* Visual identifying the entry */
    ColormapPtr cmap;			/* Colormap containing reservations */
    int nResRed, nResGreen, nResBlue;	/* How many of each (red for pseudo) */
} ColorResRec, *ColorResPtr;


/*
 * Given a screen structure and a visual ID, this macro sets the
 * last argument to the reserved color structure associated with the
 * visual (if any).
 */
#define	vReservedColors(pScrn, visid, pCRes)				\
		for (pCRes = (ColorResPtr) sResColor(pScrn);		\
		    pCRes && pCRes->vid != visid;			\
		    pCRes = pCRes->next)				\
		    ;

#define	NElements(a)	(sizeof (a) / sizeof (a)[0])

#define	SecWireEvent(n)	(SecExtension->eventBase + (n))

#define WriteSecReply(pClient, size, pReply)				\
	if ((pClient)->swapped)						\
	    (*SSecReplyVector[((xReq *)(pClient)->requestBuffer)->data]) \
		(pClient, (int)(size), pReply);				\
	else								\
	    (void) WriteToClient(pClient, (int)(size), (char *)(pReply));

#define CURRENT_REQUEST(client)	(((xReq *)(client->requestBuffer))->reqType)

#define SEC_REQUEST(minor)	((SecExtension->base << 8) + (minor))

/*
 * Access- and Privilege-check macros.
 */

#define SubjAttrPrivCheck(clntattrs, priv)				\
	(PrivIsInSet((clntattrs)->effPrivs, (priv)))

#define SubjHasPrivilege(clnt, priv, req)				     \
	(SubjAttrPrivCheck(ClientSecAttrs(clnt), (priv)) ? 		     \
	 (AuditUseOfPrivilege((clnt), 0, (priv), (req), (char *) 0), TRUE) : \
	 (AuditLackOfPrivilege((clnt), (priv), (req)), FALSE))

#define SubjHasAccess(subjattr, objattr, acctype, aud, req, id1, id2, id3) \
	(((subjattr)->clientID == (objattr)->creator) ? 0 :		\
	    FullAccessCheck((subjattr), (objattr), (acctype), (aud),	\
		(req), (id1), (id2), (id3)))

/*
 * Check a client's access to the given window.
 */
#define CHECK_WINDOW_ACCESS(client, pWin, acc, req, wid, errp)		\
	{								\
	    int access;							\
	    if ((access = SubjHasAccess(ClientSecAttrs(client),		\
		wSecAttrs(pWin), (acc), TRUE, (req),			\
		(pWin)->drawable.id, 0L, 0L)) != 0)			\
	    {								\
		(client)->errorValue = wid;				\
		*(errp) = (access < 0) ? BadWindow : BadAccess;		\
	    }								\
	    else							\
		*(errp) = 0;						\
	}

/*
 * This macro is used when the value returned in an error event should
 * not include the window ID of the affected window.  For example, if you
 * attempt a SendEvent to the current focus window and you don't have the
 * proper access, we don't want to send the window ID of the current
 * focus window in the error event.
 */
#define VERIFY_AUX_WINDOW_ACCESS(client, pWin, acc, req, wid)		\
	{								\
	    int error;							\
	    CHECK_WINDOW_ACCESS(client, pWin, acc, req, wid, &error);	\
	    if (error)							\
		return error;						\
	}

/*
 * This macro is the usual case.  Return bad status to dispatch if the
 * client doesn't have the required access.
 */
#define VERIFY_WINDOW_ACCESS(client, pWin, acc, req)			\
	VERIFY_AUX_WINDOW_ACCESS(client, pWin, acc, req,		\
		(pWin)->drawable.id)

/*
 * If the access fails and MAC-read access is denied, return an error.
 * If access is denied for some other reason, write a reply.
 */
#define VERIFY_WINDOW_ACCESS_REPLY(client, pWin, acc, req, rep, repsz)	\
	{								\
	    int error;							\
	    CHECK_WINDOW_ACCESS(client, pWin, acc, req,			\
		(pWin)->drawable.id, &error);				\
	    if (error != 0)						\
	    {								\
		if (error < 0)						\
		    return error;					\
		WriteReplyToClient(client, repsz, rep);			\
		return Success;						\
	    }								\
	}


/*
 * Same as above, except this is used when argument is drawable
 */
#define VERIFY_DRAWABLE_ACCESS(client, pDraw, acctype, req)		\
	{								\
	    ObjSecAttr *objattr;					\
	    int access;							\
	    if ((pDraw)->type == DRAWABLE_PIXMAP)			\
	    {								\
		PixmapSecRecPtr psecpmap;				\
		(void) LookupPixmap((pDraw)->id, &psecpmap);		\
		objattr = psecpmap->secAttrs;				\
	    }								\
	    else							\
		objattr = wSecAttrs((WindowPtr) pDraw);			\
	    if ((access = SubjHasAccess(ClientSecAttrs(client), (objattr), \
		(acctype), TRUE, (req), (pDraw)->id, 0L, 0L)) != 0)	\
	    {								\
		(client)->errorValue = (pDraw)->id;			\
		return (access < 0) ? BadDrawable : BadAccess;		\
	    }								\
	}

#define CHECK_PIXMAP_ACCESS(client, pPixmap, pSecPixmap, acctype, req, errp) \
	{								\
	    PixmapSecRecPtr psecpmap;					\
	    XID pid;							\
	    int access;							\
	    pid = (pPixmap)->drawable.id;				\
	    if (pSecPixmap == (PixmapSecRecPtr) 0)			\
		psecpmap = (PixmapSecRecPtr) LookupPixmap(pid, &psecpmap); \
	    else							\
		psecpmap = pSecPixmap;					\
	    if ((access = SubjHasAccess(ClientSecAttrs(client),		\
		psecpmap->secAttrs, (acctype), TRUE, (req),		\
		pid, 0L, 0L)) != 0)					\
	    {								\
		(client)->errorValue = pid;				\
		*(errp) = (access < 0) ? BadPixmap : BadAccess;		\
	    }								\
	    else							\
		*(errp) = 0;						\
	}

#define VERIFY_PIXMAP_ACCESS(client, pPixmap, pSecPixmap, acctype, req)	\
	{								\
	    int error;							\
	    CHECK_PIXMAP_ACCESS(client, pPixmap, pSecPixmap, acctype, 	\
		req, &error);						\
	    if (error)							\
		return error;						\
	}

#define CHECK_COLORMAP_ACCESS(client, pmap, psecmap, acctype, req, errp)\
	{								\
	    int access;							\
	    if ((access = SubjHasAccess(ClientSecAttrs(client),		\
		(psecmap)->secAttrs, (acctype), TRUE, (req),		\
		(pmap)->mid, 0L, 0L)) != 0)				\
	    {								\
		(client)->errorValue = (pmap)->mid;			\
		*(errp) = (access < 0) ? BadColor : BadAccess;		\
	    }								\
	    else							\
		*(errp) = 0;						\
	}

#define VERIFY_COLORMAP_ACCESS(client, pmap, psecmap, acctype, req)	\
	{								\
	    int error;							\
	    CHECK_COLORMAP_ACCESS(client, pmap, psecmap, acctype, 	\
		req, &error);						\
	    if (error)							\
		return error;						\
	}

#define VERIFY_AUX_GC_ACCESS(client, pGC, acctype, req, gid, err)	\
	{								\
	    int access;							\
	    if ((access = SubjHasAccess(ClientSecAttrs(client),		\
		gcSecAttrs(pGC), (acctype), TRUE, (req),		\
		gid, 0L, 0L)) != 0)					\
	    {								\
		(client)->errorValue = gid;				\
		return (access < 0) ? err : BadAccess;			\
	    }								\
	}

#define VERIFY_GC_ACCESS(client, pGC, acctype, req, gid)		\
	VERIFY_AUX_GC_ACCESS(client, pGC, acctype, req, gid, BadGC)

#define CHECK_CURSOR_ACCESS(client, pSecCurs, acctype, req, cid, errp)	\
	{								\
	    int access;							\
	    if ((access = SubjHasAccess(ClientSecAttrs(client),		\
		(pSecCurs)->secAttrs, (acctype), TRUE, (req),		\
		cid, 0L, 0L)) != 0)					\
	    {								\
		(client)->errorValue = cid;				\
		*(errp) = (access < 0) ? BadCursor : BadAccess;		\
	    }								\
	    else							\
		*(errp) = 0;						\
	}

#define VERIFY_CURSOR_ACCESS(client, pSecCurs, acctype, req, cid)	\
	{								\
	    int error;							\
	    CHECK_CURSOR_ACCESS(client, pSecCurs, acctype, req, cid,	\
		&error);						\
	    if (error)							\
		return error;						\
	}

/* 
 * This macro is used when doing the translation from ufid (user font ID)
 * to internal font ID.  It does the first lookup which includes the
 * access check, then does the internal font ID lookup.
 * If the user-provided font ID does not identify a font, no return
 * from the invoking routine is made (the pFont argument is set to 0).
 * If there is an access error, however, the error IS returned.
 */

#if defined(AIXV3) && defined(R4DDX)
#define	LOOKUP_VERIFY_FONT_ACCESS(client, ufid, acctype, req, pFont)	\
	{								\
	FontPtr		tmp;	/*R4 font ptr */			\
									\
	    int err;							\
	    Font ifid;							\
	    err = LookupUserFid(ufid, client, acctype, req, &ifid);	\
	    if (err != Success)						\
	    {								\
		if (ifid)		/* Access error */		\
		    return err;						\
		pFont = 0;						\
	    }								\
	    else							\
		{							\
		tmp = (FontPtr)LookupIDByType(ifid, RT_FONT);		\
		pFont = (FontPtr) R5FromR4FontPtr(tmp);			\
		}							\
	}
#else
#define	LOOKUP_VERIFY_FONT_ACCESS(client, ufid, acctype, req, pFont)	\
	{								\
	    int err;							\
	    Font ifid;							\
	    err = LookupUserFid(ufid, client, acctype, req, &ifid);	\
	    if (err != Success)						\
	    {								\
		if (ifid)		/* Access error */		\
		    return err;						\
		pFont = 0;						\
	    }								\
	    else							\
		pFont = (FontPtr)LookupIDByType(ifid, RT_FONT);		\
	}
#endif

/*
 * This macro simplifies the source a little by allowing us to pass
 * SEC_MAC-only multilevel property names to the ..._PROPERTY_ACCESS macros.
 * The macros will not actually use the variable name in non-SEC_MAC
 * configurations so there won't be an undeclared variable reference.
 */
#if SEC_MAC
# define MLAtomName(pProp, atom)	atom
#else
# define MLAtomName(pProp, atom)	(pProp)->propertyName
#endif

#define CHECK_PROPERTY_ACCESS(client, pWin, pProp, pSec, atom, acctype, req) \
	    SubjHasAccess(ClientSecAttrs(client), (pSec)->secAttrs,	\
		(acctype), TRUE, (req), (pWin)->drawable.id,		\
		MLAtomName(pProp, atom), 0L)

#define VERIFY_PROPERTY_ACCESS(client, pWin, pProp, pSec, atom, acctype, req) \
	{								\
	    if (CHECK_PROPERTY_ACCESS(client, pWin, pProp, pSec, atom,	\
		acctype, req) != 0)					\
	    {								\
		(client)->errorValue = MLAtomName(pProp, atom);		\
		return BadAccess;					\
	    }								\
	}

extern WindowPtr	*WindowTable;

#define	IsRootWindow(pWin)	(WindowTable[(pWin)->drawable.pScreen->myNum] \
				== (pWin))

#if SEC_MAC
#define AllocMACLabel()		((MACLabel *)xalloc(MAC_LABEL_SIZE_RND))
#define FreeMACLabel(l)		xfree((char *)l)
#define CopyMACLabel(src,dst)	memcpy(dst, src, MAC_LABEL_SIZE)
#endif

#if SEC_ILB
#define AllocInfoLabel()	((InfoLabel *)xalloc(INFO_LABEL_SIZE_RND))
#define FreeInfoLabel(l)	xfree((char *)l)
#define CopyInfoLabel(src,dst)	memcpy(dst, src, INFO_LABEL_SIZE)
#endif

#define AUDIT_ACCESS_DENIAL	1	/* Access to object denied */
#define AUDIT_PRIVILEGE_DENIAL	2	/* Lack of privilege */
#define AUDIT_CONNECTION	3	/* Successful connection */
#define AUDIT_NO_CONNECTION	4	/* Unsuccessful connection attempt */
#define AUDIT_DISCONNECTION	5	/* Client disconnection */

extern ExtensionEntry	*SecExtension;

extern void	(*SSecReplyVector[])();

extern int	WriteToClient();

extern Bool	IsWindowMgrWindow();
extern int	FullAccessCheck();
extern void	AuditNoPrivilege();


#endif /* } SEC_BASE */
#endif /* _XSEC_H */
